#include <linux/init.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/if_arp.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_arp.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>

#include "shadaarpko.h"


MODULE_LICENSE("GPL");   /*这行用于告诉内核该模块拥有free license,在2.6中这是必须的*/
MODULE_AUTHOR("laf163@gmail.com");

#define DEVICE_NAME	"shadaarp"

struct arpinfo
{
    __be16        ar_hrd;         /* format of hardware address   */
    __be16        ar_pro;         /* format of protocol address   */
    unsigned char ar_hln;         /* length of hardware address   */
    unsigned char ar_pln;         /* length of protocol address   */
    __be16        ar_op;          /* ARP opcode (command)         */

    unsigned char ar_sha[ETH_ALEN];   /* sender hardware address  */
    unsigned char ar_sip[4];          /* sender IP address        */
    unsigned char ar_tha[ETH_ALEN];   /* target hardware address  */
    unsigned char ar_tip[4];          /* target IP address        */

};

static struct mutex mutex;
static struct mutex mutex_attack_buf;
static struct mutex mutex_arp_buf;

static char gateway_ip[16]={0};
static char gateway_mac[18]={0};
#define ATTACK_BUF_MAX 256
#define ARP_BUF_MAX 256
//static char g_attack_buf[ATTACK_BUF_MAX]={0};//攻击数据
static char g_arp_buf[ARP_BUF_MAX]={0};//所有接收到的arp数据缓冲区
static unsigned int g_arp_count=0;//ARP包数量
static unsigned int arphookfn(unsigned int hook,
                              struct sk_buff *skb,
                              const struct net_device *in,
                              const struct net_device *out,
                              int(*okfn)(struct sk_buff *))
{
    struct arpinfo *arp=NULL;
    char ar_sha[18]={0};
    char ar_sip[16]={0};
    char ar_tha[18]={0};
    char ar_tip[16]={0};
    bool bAttack=false;

    //不是arp包，直接返回
    if (skb->protocol != htons(ETH_P_ARP))return NF_ACCEPT;
    arp=(struct arpinfo*)arp_hdr(skb);

    sprintf(ar_sha,"%02x:%02x:%02x:%02x:%02x:%02x",
            arp->ar_sha[0],arp->ar_sha[1],arp->ar_sha[2],
            arp->ar_sha[3],arp->ar_sha[4],arp->ar_sha[5]);
    sprintf(ar_sip,"%d.%d.%d.%d",
            arp->ar_sip[0],arp->ar_sip[1],
            arp->ar_sip[2],arp->ar_sip[3]);
    sprintf(ar_tha,"%02x:%02x:%02x:%02x:%02x:%02x",
            arp->ar_tha[0],arp->ar_tha[1],arp->ar_tha[2],
            arp->ar_tha[3],arp->ar_tha[4],arp->ar_tha[5]);
    sprintf(ar_tip,"%d.%d.%d.%d",
            arp->ar_tip[0],arp->ar_tip[1],
            arp->ar_tip[2],arp->ar_tip[3]);

    if (arp->ar_op==htons(ARPOP_REPLY)) //arp应答包
    {
        //printk(KERN_ALERT "\narp reply");
        mutex_lock(&mutex);
        if (strcmp(ar_sip,gateway_ip)==0)
        {
            if (strcmp(ar_sha,gateway_mac)!=0) //发现攻击
            {
                //printk(KERN_ALERT "\n Found arp attack!!!");
                bAttack=true;
            }
        }
        mutex_unlock(&mutex);
    }
    //else if (arp->ar_op==htons(ARPOP_REQUEST))
    //{
        //printk(KERN_ALERT "\narp request");
    //}

    //保存arp，供应用程序读取
    //格式：
    //发送者mac\n
    //发送者IP\n
    //目标mac\n
    //目标ip\n
    //操作码\n
    //类型：0为正常,1为攻击\n
    if ((ARP_BUF_MAX-strlen(g_arp_buf))>(strlen(ar_sip)+strlen(ar_sha)+strlen(ar_tip)+strlen(ar_tha)+7))
    {
        mutex_lock(&mutex_arp_buf);
        strcat(g_arp_buf,ar_sha);
        strcat(g_arp_buf,"\n");
        strcat(g_arp_buf,ar_sip);
        strcat(g_arp_buf,"\n");
        strcat(g_arp_buf,ar_tha);
        strcat(g_arp_buf,"\n");
        strcat(g_arp_buf,ar_tip);
        strcat(g_arp_buf,"\n");
        strcat(g_arp_buf,arp->ar_op==htons(ARPOP_REQUEST)?"1":"2");
        strcat(g_arp_buf,"\n");
        strcat(g_arp_buf,bAttack?"1\n":"0\n");
        g_arp_count++;
        mutex_unlock(&mutex_arp_buf);
    }
    if(bAttack)return NF_DROP;
    return NF_ACCEPT;//允许地址解析协议报

}

static int shadaarp_open(struct inode *inode, struct file *filep)
{
    if (!try_module_get(THIS_MODULE))return -EBUSY;
    return 0;
}

static int shadaarp_release(struct inode *inode, struct file *filep)
{
    module_put(THIS_MODULE);
    return 0;
}

static int shadaarp_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int cmd,
    unsigned long arg)
{
    struct ioctl_data data;
    struct ioctl_data *pdata;
    int l=0;
    unsigned int arp_count=0;

    pdata=(struct ioctl_data*)arg;
    switch (cmd)
    {
    case COMMAND_SET_GATEWAY:
        if (copy_from_user(&data,pdata,sizeof(struct ioctl_data))!=0)
            return -EINVAL;
        printk(KERN_ALERT "\ngateway ip =%s",data.ip);
        printk(KERN_ALERT "\ngateway mac=%s",data.mac);
        mutex_lock(&mutex);
        strcpy(gateway_ip,data.ip);
        strcpy(gateway_mac,data.mac);
        //strlwr(gateway_mac);//MAC转为小写
        mutex_unlock(&mutex);
        break;
    /*case COMMAND_GET_ATTACK:
        mutex_lock(&mutex_attack_buf);
        l=strlen(g_attack_buf);
        if (copy_to_user((char*)arg,g_attack_buf, l+1))
        {
            mutex_unlock(&mutex_attack_buf);
            return -EFAULT;
        }
        g_attack_buf[0]=0;
        mutex_unlock(&mutex_attack_buf);
        return l;*/
    case COMMAND_GET_ARP:
        if (g_arp_count>0)
        {
            mutex_lock(&mutex_arp_buf);
            l=strlen(g_arp_buf);
            if (copy_to_user((char*)arg,g_arp_buf, l+1))
            {
                mutex_unlock(&mutex_arp_buf);
                return -EFAULT;
            }
            g_arp_buf[0]=0;
            arp_count=g_arp_count;
            g_arp_count=0;
            mutex_unlock(&mutex_arp_buf);
            return arp_count;
        }
        break;
    default:
        return -EINVAL;
    }

    return 0;
}

static struct file_operations fileops=
{
    .llseek         = NULL,
                      .read           = NULL,
                                        .write          = NULL,
                                                          .poll           = NULL,
                                                                            .ioctl          = shadaarp_ioctl,
                                                                                              .compat_ioctl   = NULL,
                                                                                                                .open           = shadaarp_open,
                                                                                                                                  .release        = shadaarp_release,
                                                                                                                                                    .fasync         = NULL,
                                                                                                                                                                  };
static struct cdev *dev;
static dev_t devt;
static struct class *shadaarp_class;
static struct nf_hook_ops ops[22]=
{
    {
        {
            NULL,NULL
        },
        arphookfn,
        THIS_MODULE,
        NF_ARP,
        NF_ARP_OUT,
        NF_IP_PRI_FILTER
    },
    {
        {NULL,NULL},
        arphookfn,
        THIS_MODULE,
        NF_ARP,
        NF_ARP_IN,
        0
    },
};

/*初始化*/
static int shadaarp_init(void)
{
    int ret;
    int major;

    /*申请主设备号*/
    ret=alloc_chrdev_region(&devt,0,1,DEVICE_NAME);
    if (ret<0)
    {
        printk(DEVICE_NAME "\ncan't register major number");
        return ret;
    }
    major = MAJOR(devt);
    printk(KERN_ALERT "\nmajor=%d",major);
    dev=cdev_alloc();
    if (dev==NULL)
    {
        printk(DEVICE_NAME "\ncdev_alloc error");
        return -1;
    }
    cdev_init(dev,&fileops);
    //dev->owner = THIS_MODULE;
    //dev->ops=&fileops;
    ret = cdev_add (dev,devt,1);
    if (ret<0)
    {
        printk(DEVICE_NAME "\ncdev_add error");
        return -1;
    }
    shadaarp_class=class_create(THIS_MODULE,"shada_class1");
    if (IS_ERR(shadaarp_class))
    {
        printk(KERN_ALERT "\nCreate class error");
        return -1;
    }
    device_create(shadaarp_class,NULL,devt,NULL,"shadaarp");
    if (nf_register_hooks(ops,2)!=0)
    {
        printk(KERN_ALERT "\nshadaarp: nf_register_hook() return !0.");
        printk(KERN_ALERT "\nshadaarp: init fail.");
        return -1;
    }
    printk(KERN_ALERT "\nshadaarp: init successful.");
    printk(KERN_ALERT "\nshadaarp: the process is \"%s\" (pid %i)", current->comm, current->pid);

    mutex_init(&mutex);
    mutex_init(&mutex_attack_buf);
    mutex_init(&mutex_arp_buf);
    return 0;
}

/*卸载*/
static void shadaarp_exit(void)
{
    printk(KERN_ALERT "\nshadaarp: unload\n");
    nf_unregister_hooks(ops,2);
    cdev_del(dev);
    device_destroy(shadaarp_class,devt);
    class_destroy(shadaarp_class);
    mutex_destroy(&mutex);
    mutex_destroy(&mutex_attack_buf);
    mutex_destroy(&mutex_arp_buf);
}

/*该函数注册模块的构造函数*/
module_init(shadaarp_init);
/*该函数注册模块的析构函数*/
module_exit(shadaarp_exit);

